/*----------------------------------------------------------------------------

  Copyright (c) 2017 rafael grompone von gioi <grompone@gmail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.

  ----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "iio.h"
//#include "itos.c" 

/*----------------------------------------------------------------------------*/
#ifndef FALSE
#define FALSE 0
#endif /* !FALSE */

#ifndef TRUE
#define TRUE 1
#endif /* !TRUE */

/*----------------------------------------------------------------------------*/
/*
   Fatal error, print a message to standard-error output and exit.
 */
static void error(char * msg)
{
  fprintf(stderr,"Error: %s\n",msg);
  exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/
/** Memory allocation, print an error and exit if fail.
 */
static void * xmalloc(size_t size)
{
  void * p;
  if( size == 0 ) error("xmalloc: zero size");
  p = malloc(size);
  if( p == NULL ) error("xmalloc: out of memory");
  return p;
}

/*----------------------------------------------------------------------------*/
/*
   Computes the logarithm of NFA to base 10.

   NFA = NT.b(n,k,p)
   the return value is log10(NFA)

   n,k,p - binomial parameters.
   logNT - logarithm of Number of Tests
 */
#define TABSIZE 100000
static double log_nfa(int n, int k, double p, double logNT)
{
  static double inv[TABSIZE];   /* table to keep computed inverse values */
  double tolerance = 0.1;       /* an error of 10% in the result is accepted */
  double log1term,term,bin_term,mult_term,bin_tail,err;
  double p_term = p / (1.0-p);
  int i;

  if( n<0 || k<0 || k>n || p<0.0 || p>1.0 )
    error("wrong n, k or p values in nfa()");

  if( n==0 || k==0 ) return logNT;
  if( n==k ) return logNT + (double)n * log10(p);

  log1term = lgamma((double)n+1.0) - lgamma((double)k+1.0)
           - lgamma((double)(n-k)+1.0)
           + (double)k * log(p) + (double)(n-k) * log(1.0-p);

  term = exp(log1term);
  if( term == 0.0 )                        /* the first term is almost zero */
    {
      if( (double)k > (double)n * p )      /* at begining or end of the tail? */
        return log1term / M_LN10 + logNT;  /* end: use just the first term */
      else
        return logNT;                      /* begin: the tail is roughly 1 */
    }

  bin_tail = term;
  for(i=k+1;i<=n;i++)
    {
      bin_term = (double)(n-i+1) * ( i<TABSIZE ?
                   (inv[i] ? inv[i] : (inv[i]=1.0/(double)i)) : 1.0/(double)i );
      mult_term = bin_term * p_term;
      term *= mult_term;
      bin_tail += term;
      if(bin_term<1.0)
        {
          /* when bin_term<1 then mult_term_j<mult_term_i for j>i.
             then, the error on the binomial tail when truncated at
             the i term can be bounded by a geometric serie of form
             term_i * sum mult_term_i^j.                            */
          err = term * ( ( 1.0 - pow( mult_term, (double)(n-i+1) ) ) /
                         (1.0-mult_term) - 1.0 );

          /* one wants an error at most of tolerance*final_result, or:
             tolerance * abs(-log10(bin_tail)-logNT).
             now, the error that can be accepted on bin_tail is
             given by tolerance*final_result divided by the derivative
             of -log10(x) when x=bin_tail. that is:
             tolerance * abs(-log10(bin_tail)-logNT) / (1/bin_tail)
             finally, we truncate the tail if the error is less than:
             tolerance * abs(-log10(bin_tail)-logNT) * bin_tail        */
          if( err < tolerance * fabs(-log10(bin_tail)-logNT) * bin_tail ) break;
        }
    }
  return log10(bin_tail) + logNT;
}

/*----------------------------------------------------------------------------*/
/* Damian's add : get the JPEG coordinate of subblock and compute the forgery map */
void get_jpeg_coord( double * image_gray,int blk_sz, double alpha,int theta, int X, int Y, int * coord_star){
	int N = blk_sz ;
	double * image = image_gray ;
 	int x,y,c,i,j,k, Bx, By;
 	double * cross_diff; 
    double * cumulated;
    int xmax = -1;
    int ymax = -1;
    double max, min ;

	cross_diff = (double *) xmalloc( X * Y * sizeof(double) );
    cumulated = (double *) xmalloc( N * N * sizeof(double) );
  /* compute absolute value of cross-difference */
  for(x=0; x<(X-1); x++)
    for(y=0; y<(Y-1); y++){
      cross_diff[x+y*X] = fabs( image[x+y*X] + image[x+1+(y+1)*X]
                              - image[x+1+y*X] - image[x+(y+1)*X] );
  if(cross_diff[x+y*X] > theta)
    cross_diff[x+y*X] = alpha * cross_diff[x+y*X] ;
    else
    cross_diff[x+y*X] = (1 - alpha) * cross_diff[x+y*X] ;
	}
  /* compute cumulated cross-difference modulo block position */
  for(x=0; x<N; x++)         /* initialization */
    for(y=0; y<N; y++)
      cumulated[x+y*N] = 0.0;
  Bx = (X-1) / N; /* number of blocks to consider */
  By = (Y-1) / N; /* (the last column and row don't have a cross-difference
                      value beacause forward difference are used. only full
                      blocks without those row and column are considered) */
  for(i=0; i<Bx; i++)        /* cumulate values */
    for(j=0; j<By; j++)
      for(x=0; x<N; x++)
        for(y=0; y<N; y++)
          cumulated[x+y*N] += cross_diff[ N*i+x + (N*j+y) * X ];

  /* find x of maximal column */
  max = 0.0;
  for(x=0; x<N; x++)
    {
      double v = 0.0;
      for(y=0; y<N; y++) v += cumulated[x+y*N];
      if( v > max )
        {
          max = v;
          xmax = x;
        }
    }

  /* find y of maximal row */
  max = 0.0;
  for(y=0; y<N; y++)
    {
      double v = 0.0;
      for(x=0; x<N; x++) v += cumulated[x+y*N];
      if( v > max )
        {
          max = v;
          ymax = y;
        }
    }

  /* compute number of favorable cases for maximal position */
  k = 0;
  for(i=0; i<N; i++)
    {
      if( cumulated[i+ymax*N] > cumulated[ i + ((ymax-1)%N) * N] ) ++k;
      if( cumulated[i+ymax*N] > cumulated[ i + ((ymax+1)%N) * N] ) ++k;
      if( cumulated[xmax+i*N] > cumulated[ (xmax-1)%N  +  i * N] ) ++k;
      if( cumulated[xmax+i*N] > cumulated[ (xmax+1)%N  +  i * N] ) ++k;
    }
/* return the result */
coord_star[1] = (xmax+1)%N ;
coord_star[2] = (ymax+1)%N ;
free( (void *) cross_diff );
free( (void *) cumulated );
}
/*----------------------------------------------------------------------------*/
/*                                    Main                                    */
/*----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
  double * input;
  double * image;
  double * block_ij ;
  double * block_tmp ;
  double * RDM ;
  char str_ij [20] ;
  int * coord_star ;
  int * coord_ij ; 
  int X,Y,C;
  int x,y,c,i,j,k,Bx,By;
  double max,min,lnfa, alpha ;
  int N, theta, macro_blk_sz, n_row, n_col, rmn_row, rmn_col ;

 //read input
  if( argc != 7 )
    error("use: jpegblocks <block_size> <image> <detection map> <alpha> <theta> <macroblocksize>");           
  N = atoi(argv[1]);           
  alpha = atof(argv[4]) ;
  theta = atoi(argv[5]) ;
  macro_blk_sz = 8*atoi(argv[6]) ; 
 



  printf("alpha = %f\ntheta = %d\nJPEG block size = %d\nMacro block size =%d\n", alpha, theta, N,macro_blk_sz) ;
  input = iio_read_image_double_split(argv[2], &X, &Y, &C);

  /* get memory */
  image = (double *) xmalloc( X * Y * sizeof(double) );
  n_col = floor(X/macro_blk_sz) ;
  rmn_row = X%macro_blk_sz ;
  n_row = floor(Y/macro_blk_sz) ;
  rmn_row = X%macro_blk_sz ;
  RDM = (double *) xmalloc( n_row * n_col * sizeof(double) );
  coord_star = (int *) xmalloc( 2 * sizeof(int) ) ; 
  block_ij = (double*) xmalloc( macro_blk_sz * macro_blk_sz * sizeof(double) );
  block_tmp = (double*) xmalloc( macro_blk_sz * macro_blk_sz * sizeof(double) );
  coord_ij = (int *) xmalloc( 2 * sizeof(int) ) ;
  /* generate a gray level image from the input
     (this is useful when the input has many channels) */
  for(x=0; x<X; x++)
    for(y=0; y<Y; y++)
      {
        image[x+y*X] = 0.0;
        for(c=0; c<C; c++)
          if(c==0)
          image[x+y*X] += 0.2989*input[x+y*X+c*X*Y];
          else if(c==1)
          image[x+y*X] += 0.5870*input[x+y*X+c*X*Y];
          else if(c==2)
          image[x+y*X] += 0.1140*input[x+y*X+c*X*Y];
   }
   // Divide image into sub-blocks
   get_jpeg_coord(image, N, alpha, theta,X,Y, coord_star) ;
   for(i=0;i<n_col;i++){
      for(j=0;j<n_row;j++){
       	   for(int m=0;m<macro_blk_sz*macro_blk_sz;m++){
       	          block_ij[m] = image[j*X*macro_blk_sz + i*macro_blk_sz + (int)floor(m/macro_blk_sz) +
       	          (int)(m%macro_blk_sz)*X] ; 
       	   }
 /*      	   for(int m=0;m<macro_blk_sz*macro_blk_sz;m++){
       	          block_ij[m] = block_tmp[macro_blk_sz*macro_blk_sz-m] ; 
       	   }
*/
           get_jpeg_coord(block_ij, N, alpha, theta, macro_blk_sz, macro_blk_sz,coord_ij) ;
		//   sprintf(str_ij, "block_%d%d.png",i,j) ;
         //  iio_write_image_double_split(str_ij,block_ij,macro_blk_sz,macro_blk_sz,1);
        //   printf("x_ij = %d\ny_ij = %d\n", coord_ij[2], coord_ij[1]) ;
         if((coord_ij[2]!=coord_star[1])||(coord_ij[1]!=coord_star[2]))
            RDM[i + j * n_col] = 255;
         else
            RDM[i + j * n_col] = 0;    
       }
   }
  
  printf("x_star = %d\ny_start = %d\n", coord_star[1], coord_star[2]) ;
  iio_write_image_double_split(argv[3],RDM,n_col, n_row,1);
  /* free memory */
  free( (void *) input );
  free( (void *) image );
  free( (void *) block_ij) ;
  free( (void *) coord_ij) ;  
  free( (void *) RDM) ;  

  return EXIT_SUCCESS;
}
/*----------------------------------------------------------------------------*/
